from django.db import models
from rest import models as rm


class FeedBase(models.Model):
    class Meta:
        abstract = True

    class RestMeta:
        GRAPHS = {
            "default": {
                "recurse_into": ["generic__component"]
            }
        }

    created = models.DateTimeField(auto_now_add=True, editable=True, db_index=True)
    component = models.SlugField(max_length=124, null=True, blank=True, default=None, db_index=True)
    component_id = models.IntegerField(null=True, blank=True, default=None, db_index=True)

    kind = models.SlugField(max_length=32, db_index=True)
    note = models.TextField(null=True, blank=True, default=None)

    @classmethod
    def log(cls, member, group, kind, component=None, component_id=None, note=None):
        obj = cls(member=member, group=group, kind=kind, component=component, component_id=component_id, note=note)
        obj.save()
        return obj


class MemberFeed(FeedBase, rm.RestModel):
    member = models.ForeignKey("account.Member", default=None, null=True, related_name="feed", on_delete=models.CASCADE)
    group = models.ForeignKey("account.Group", default=None, null=True, related_name="+", on_delete=models.CASCADE)


class GroupFeed(FeedBase, rm.RestModel):
    class RestMeta:
        AUTO_SAVE = ["member"]
        SEARCH_FIELDS = ["note"]
        VIEW_PERMS = ["manage_groups", "manage_group", "view_all_groups"]
        GRAPHS = {
            "default": {
                "graphs": {
                    "member": "basic",
                    "generic__component": "basic"
                }
            }
        }
    member = models.ForeignKey("account.Member", default=None, null=True, related_name="+", on_delete=models.CASCADE)
    group = models.ForeignKey("account.Group", default=None, null=True, related_name="feed", on_delete=models.CASCADE)

    def on_rest_created(self, request):
        self.group.sendEvent("chat_message", None, custom=dict(pk=self.pk, by=request.member.id, kind=self.kind))




from . import model as ms
from django.db.models.query import QuerySet
from django.db.models import Count, F, Q
from rest import settings
from datetime import datetime, timedelta
# from . import profiler
from . import csv


def serialize(qset, graph, sort=None, size=25, start=0, format="json"):
    if isinstance(graph, str):
        model = getattr(qset, "model", None)
        if model:
            graph = model.getGraph(graph)
    if format == "json":
        return to_list_from_graph(qset, graph, sort, size, start)
    return to_format(qset, format, size=size).data


def to_format(qset, format, size=10000):
    model = getattr(qset, "model", None)
    fields = model.getRestFormatFields(format)
    if format == "csv":
        return csv.generateCSV(qset, fields, f"serialize.{format}")
    raise Exception("invalid format")


def to_list_from_graph(qset, graph, sort=None, size=25, start=0):
    fields = graph.get("fields", [])
    extra = graph.get("extra", [])
    exclude = graph.get("exclude", [])
    recurse_into = graph.get("recurse_into", [])
    return to_list(qset, sort, size, fields=fields, extra=extra, exclude=exclude, recurse_into=recurse_into)


def sort_nulls_last(qset, sort_args, nulls_last_fields):
    if len(sort_args) == 1 and isinstance(nulls_last_fields, list):
        arg = sort_args[0]
        if arg.startswith("-"):
            arg = arg[1:]
        if arg not in nulls_last_fields:
            return qset.order_by(*sort_args)
    new_args = []
    for arg in sort_args:
        if arg.startswith("-"):
            new_args.append(F(arg[1:]).desc(nulls_last=True))
        else:
            new_args.append(F(arg).asc(nulls_last=True))
    return qset.order_by(*new_args)


def sort_list(qset, sort):
    qset, sort_args = get_sort_args(qset, sort)
    nulls_last = False
    if hasattr(qset, "model") and hasattr(qset.model, "RestMeta"):
        nulls_last = getattr(qset.model.RestMeta, "NULLS_LAST", False)
    try:
        if nulls_last:
            qset = sort_nulls_last(qset, sort_args, nulls_last)
        else:
            qset = qset.order_by(*sort_args)
    except Exception as err:
        return qset, str(err)
    return qset, sort_args


def to_list(qset, sort=None, size=25, start=0, fields=[], extra=[], exclude=[], recurse_into=[], cache=None):
    if cache is None and isinstance(qset, QuerySet):
        cache = dict()
    output = {"size": size, "start": start}
    if sort and isinstance(qset, QuerySet):
        qset, sort_args = sort_list(qset, sort)
        output["sort"] = sort_args

    qset = qset[start:start+size]
    if not fields:
        fields = ms.get_fields(qset)
    if settings.REST_SELECT_RELATED:
        # this should improve speed greatly for lookups
        foreign_fields = ms.get_select_related_fields(qset.model, fields)
        if foreign_fields:
            qset = qset.select_related(*foreign_fields)
    data = []
    for obj in qset:
        data.append(ms.to_dict(obj, fields, extra, exclude, recurse_into, cache=cache))
    output["count"] = len(data)
    output["data"] = data
    return output


def convert_date_sort_to_dt(arg):
    v = int(arg[:-1])
    if arg.endswith("d"):
        return datetime.now() - timedelta(days=v)
    if arg.endswith("h"):
        return datetime.now() - timedelta(hours=v)
    if arg.endswith("m"):
        return datetime.now() - timedelta(minutes=v)
    return datetime.now()


def get_sort_args(qset, sort):
    if not isinstance(sort, str) or "metadata" in sort:
        return qset, None
    sort_args = []
    for s in sort.split(","):
        if s.endswith("_display"):
            # fix for django _display kinds being sorted
            s = s[:s.find("_display")]
        s = s.replace('.', '__')
        if "__count" in s:
            fields = s.split("__")
            k = fields[0]
            j = fields[1]
            s = "fk_count"
            if k.startswith("-"):
                k = k[1:]
                s = "-fk_count"

            if len(fields) > 2 and hasattr(qset.model, "RestMeta"):
                # this is a daterange field, 5m minutes, 2h hours, 40d days
                # the daterange field is the DATE_RANGE_FIELD
                dr_field = getattr(qset.model.RestMeta, "DATE_RANGE_FIELD", "created")
                dr = fields[2]
                q = {}
                q[f"{k}__{dr_field}__gte"] = convert_date_sort_to_dt(dr)
                qfilter = Q(**q)
                qset = qset.annotate(fk_count=Count(k, filter=qfilter)).distinct()
            else:
                qset = qset.annotate(fk_count=Count(k)).distinct()
        sort_args.append(s)
    return qset, sort_args



from .rules import *  # noqa: F401, F403
from .incident import *  # noqa: F401, F403
from .event import *  # noqa: F401, F403
from .ossec import *  # noqa: F401, F403
from .ticket import *  # noqa: F401, F403



#-*- coding: utf-8 -*-
from socket import error as socket_error
from django.http import BadHeaderError


class WebSocketError(socket_error):
    """
    Raised when an active websocket encounters a problem.
    """


class FrameTooLargeException(WebSocketError):
    """
    Raised if a received frame is too large.
    """


class HandshakeError(BadHeaderError):
    """
    Raised if an error occurs during protocol handshake.
    """


class UpgradeRequiredError(HandshakeError):
    """
    Raised if protocol must be upgraded.
    """

class SSLRequiredError(socket_error):
    """
    Raised if protocol must be upgraded.
    """



# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility module to import a JSON module

Hides all the messy details of exactly where
we get a simplejson module from.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'


try: # pragma: no cover
  # Should work for Python2.6 and higher.
  import json as simplejson
except ImportError: # pragma: no cover
  try:
    import simplejson
  except ImportError:
    # Try to import from django, should work on App Engine
    from django.utils import simplejson



import inspect
import json
import time
import datetime
import math
from decimal import Decimal


def rest_serialize(data):
    if isinstance(data, datetime.datetime):
        try:
            return time.mktime(data.timetuple())
        except Exception:
            print(("invalid date: {0}".format(data)))
        return 0.0
    elif isinstance(data, datetime.date):
        return data.strftime("%Y/%m/%d")
    elif isinstance(data, dict):
        data = data.copy()
        for k in list(data.keys()):
            data[k] = rest_serialize(data[k])
    elif isinstance(data, list) or isinstance(data, tuple):
        newdata = []
        for v in data:
            newdata.append(rest_serialize(v))
        data = newdata
    elif isinstance(data, float):
        if math.isnan(data):
            return 0.0
    elif isinstance(data, Decimal):
        if data.is_nan():
            return 0.0
        return float(data)
    return data


def requestInPaths(request, paths):
    for ep in paths:
        if request.path.startswith(ep):
            return True
    return False


def _getFields(qset=None, model=None):
    """
    returns a list of fields for specified QuerySet or Model
    """
    if hasattr(qset, '_meta') and getattr(qset, '_meta') != None:
        fields = qset._meta.fields
    elif hasattr(qset, 'model') and getattr(qset, 'model') != None:
        fields = qset.model._meta.fields
    elif hasattr(qset, 'keys'):
        return list(qset.keys())
    elif hasattr(model, '_meta') and getattr(model, '_meta') != None:
        fields = model._meta.fields
    else:
        return None
    if fields:
        return list(f.name for f in fields)
    return None


def _filter_recurse(remove, lst):
    if isinstance(lst, dict):
        ret = {}
    else:
        ret = []
    for item in lst:
        if isinstance(lst, dict):
            if item[:len(remove)+1] == remove + ".":
                ret[item[len(remove)+1:]] = lst[item]
        elif type(item) in (list,tuple) and type(item[0]) in (str, str):
            p = item[0]
            if p[:len(remove)+1] == remove + ".":
                ret.append((p[len(remove)+1:], item[1]))
        elif isinstance(item, str):
            if item[:len(remove)+1] == remove + ".":
                ret.append(item[len(remove)+1:])

    return ret


def __call_func(func, *args, **kwargs):
    if not kwargs:
        return func(*args)
    if hasattr(inspect, "getargspec"):
        # removed in python3.11 (but much faster)
        take = inspect.getargspec(func)[0]
    else:
        take = list(inspect.signature(func).parameters.keys())
    give = {}
    for arg in kwargs:
        if arg in take:
            give[arg] = kwargs[arg]
    return func(*args, **give)


from medialib.stores.s3store import getS3
from rest.helpers import toBytes
from django.conf import settings
from datetime import datetime
import boto3
from rest import helpers as rest_helpers

def UPLOAD(task):
    when = rest_helpers.parseDate(task.data.when)
    if not when:
        when = datetime.now()
    bucket = task.data.bucket
    filename = task.data.filename
    folder = task.data.folder
    key = task.data.aws
    secret = task.data.secret
    data = toBytes(task.data.data)
    if not key and not secret:
        key = settings.AWS_KEY
        secret = settings.AWS_SECRET
    if filename and '{date' in filename:
        filename = filename.format(date=when)
    if folder:
        filename = "{}/{}".format(folder.strip('/'), filename)
    try:
        s3 = boto3.resource('s3', aws_access_key_id=key, aws_secret_access_key=secret)
        s3_obj = s3.Object(bucket, filename)
        result = s3_obj.put(Body=data)
    except Exception as err:
        task.failed(err)
        return False
    task.log("data written to bucket {} at {}".format(bucket, filename))
    return True


def normalizePath(bucket, filename):
    if "/" in bucket:
        bucket = bucket.strip('/')
        pts = bucket.split('/')
        bucket = pts.pop(0)
        path = "/".join(pts)
        filename = f"{path}/{filename}"
    return bucket, filename


def uploadBody(filename, body, bucket, key, secret):
    bucket, filename = normalizePath(bucket, filename)
    s3 = boto3.resource('s3', aws_access_key_id=key, aws_secret_access_key=secret)
    s3_obj = s3.Object(bucket, filename)
    body = toBytes(body)
    return s3_obj.put(Body=body)


def uploadFile(filename, fp, bucket, key, secret):
    bucket, filename = normalizePath(bucket, filename)
    s3 = boto3.resource('s3', aws_access_key_id=key, aws_secret_access_key=secret)
    s3_obj = s3.Object(bucket, filename)
    return s3_obj.upload_fileobj(fp)



from rest import decorators as rd
from rest import views as rv
from pushit.models import Product, Release


@rd.url('product')
@rd.url('product/<str:pk>')
def rest_on_product(request, pk=None):
    if pk is not None:
        if isinstance(pk, str) and not pk.isdigit():
            obj = Product.objects.filter(oid=pk).last()
            if obj is not None:
                if request.method == "GET":
                    return obj.on_rest_get(request)
                elif request.method == "POST":
                    return obj.on_rest_post(request)
    return Product.on_rest_request(request, pk)


@rd.url('release')
@rd.url('release/<int:pk>')
def rest_on_release(request, pk=None):
    return Release.on_rest_request(request, pk)




from django.db import models
from rest import models as rm
from rest import fields as rf
from rest import settings
from objict import objict
from datetime import datetime

CM_BACKENDS = objict()
DEVICE_USE_BUID = settings.get("DEVICE_USE_BUID", False)


class MemberDevice(models.Model, rm.RestModel, rm.MetaDataModel):
    """
    MemberDevice Model tracks personal devices associated with a user.
    This can include mobile and desktop devices.
    """
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True, db_index=True)

    member = models.ForeignKey("account.Member", related_name="devices", on_delete=models.CASCADE)

    name = models.CharField(max_length=128, blank=True, null=True, default=None)
    uuid = models.CharField(db_index=True, max_length=128, blank=True, null=True, default=None)
    buid = models.CharField(db_index=True, max_length=128, blank=True, null=True, default=None)

    cm_provider = models.CharField(db_index=True, max_length=64, default="fcm")
    cm_token = models.CharField(max_length=250, default=None, null=True)

    kind = models.CharField(db_index=True, max_length=64, default="unknown")
    ip = models.CharField(max_length=64, default=None, null=True)
    state = models.IntegerField(db_index=True, default=1)

    class RestMeta:
        GRAPHS = {
            "default": {
                "extra": ["metadata"],
            }
        }

    def sendData(self, message, **kwargs):
        messenger = getCloudMessenger(self.cm_provider)
        if messenger:
            resp = messenger.sendToDevice(self, message)
            return resp
        return objict(status_code=404, reason=self.cm_provider)

    def sendNotification(self, title, body):
        messenger = getCloudMessenger(self.cm_provider)
        if messenger:
            resp = messenger.sendNotification(self.cm_token, title, body)
            return resp
        return objict(status_code=404, reason=self.cm_provider)

    def notify(self, title, body):
        return self.sendNotification(title, body)

    def touch(self, ip=None):
        if ip is not None:
            self.ip = ip
        self.modified = datetime.now()
        self.save()

    @classmethod
    def sendMessageTo(cls, message, devices, **kwargs):
        pass

    @classmethod
    def register(cls, request, member, device_id):
        cm_token = request.DATA.get("cm_token")
        default_provider = "ws"
        if cm_token is not None:
            default_provider = "fcm"
        cm_provider = request.DATA.get("cm_provider", default=default_provider)
        buid = request.DATA.get("__buid__", None)
        md = MemberDevice.objects.filter(uuid=device_id, member=member).last()
        if md is not None:
            if md.cm_token != cm_token:
                md.cm_token = cm_token
            if buid and md.buid != buid:
                md.buid = buid
            if cm_provider is not None:
                md.cm_provider = cm_provider
            if md.state != 1:
                md.state = 1
            md.touch(request.ip)
            metadata = request.DATA.get("device_metadata", None)
            if isinstance(metadata, dict):
                md.set_metadata(metadata)
            return md
        md = MemberDevice(
            uuid=device_id, buid=buid,
            member=member, cm_token=cm_token,
            cm_provider=cm_provider)
        md.ip = request.ip
        md.name = F"{member.first_name} {request.auth_session.device}"
        md.kind = request.auth_session.os.lower()
        if md.kind.startswith("mac"):
            md.kind = "mac"
        md.save()
        md.setProperty("user_agent", request.META.get('HTTP_USER_AGENT', ''))
        metadata = request.DATA.get("device_metadata", None)
        if isinstance(metadata, dict):
            md.set_metadata(metadata)
        return md

    @classmethod
    def unregister(cls, member, device_id):
        md = MemberDevice.objects.filter(uuid=device_id, member=member, state=1).last()
        if md is None:
            return False
        md.cm_token = None
        md.state = -1
        md.save()
        return True


class MemberDeviceMetaData(rm.MetaDataBase):
    parent = models.ForeignKey(MemberDevice, related_name="properties", on_delete=models.CASCADE)


class CloudCredentials(models.Model, rm.RestModel, rm.MetaDataModel):
    """
    CloudCredentials is a global setting for a group to store this groups cloud credentials.
    """
    class RestMeta:
        VIEW_PERMS = ["view_cm", "manage_cm"]
        EDIT_PERMS = ["manage_cm"]
        GRAPHS = {
            "default": {
                "extra": ["metadata"],
                "graphs": {
                    "group": "basic"
                }
            }
        }

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    group = models.ForeignKey(
        "account.Group", null=True, default=None,
        related_name="cloud_credentials", on_delete=models.CASCADE)

    name = models.CharField(max_length=128, blank=True, null=True, default=None)
    uuid = models.CharField(db_index=True, max_length=64, blank=True, null=True, default=None)
    state = models.IntegerField(db_index=True, default=1)

    # specify the cloud provider class "account.fcm.v1.FirebaseNotifier"
    # provider_class = models.CharField(max_length=255, null=True, default=None)

    credentials = rf.JSONField()

    _notifier = None

    @property
    def notifier(self):
        if self._notifier is None:
            from account import fcm
            self._notifier = fcm.v1.FirebaseNotifier(self.credentials) 
        return self._notifier

    def sendToDevice(self, device, message):
        return self.notifier.send(device.cm_token, None, None, message)

    def sendNotification(self, token, title, body):
        return self.notifier.send(token, title, body)


class CloudCredentialsMetaData(rm.MetaDataBase):
    parent = models.ForeignKey(CloudCredentials, related_name="properties", on_delete=models.CASCADE)


def getCloudMessenger(name):
    creds = CloudCredentials.objects.filter(uuid=name).last()
    if creds is not None:
        return creds
    return None



"""
This is where you can put handlers for running async background tasks

Task.Publish("mymo", "on_background_job")
"""
import time
from rest.log import getLogger
from medialib.models import MediaItem

logger = getLogger("medialib", filename="medialib.log")

# this is a background job handler
def on_render(task):
    logger.info("media task called", task.data)
    try:
        item = MediaItem.objects.get(id=task.data.id)
        item.createRenditions()
        task.completed()
    except Exception as err:
        task.failed(str(err))




import io
from rest import settings
from medialib.stores import s3
from django.http import HttpResponse, StreamingHttpResponse
from datetime import datetime
AWS_CACHING_BUCKET = settings.get("AWS_CACHING_BUCKET", settings.AWS_S3_BUCKET)


def on_cached_rest_list(Model, request):
    cache_guid = request.DATA.get("format_filename")
    file_format = request.DATA.get("format", None)
    if not file_format:
        file_format = "csv"
        request.DATA.set("format", file_format)
    content_type = "text/csv"
    if file_format[:3] == "csv":
        file_format = "csv"
    else:
        file_format = "json"
        content_type = "application/json"
    if cache_guid:
        resp = getCachedResponse(cache_guid)
        if resp is not None:
            return resp
        resp = Model.on_rest_list(request)
        # TODO only cache if dr_end exists and is less then today
        dr_end = request.DATA.get("dr_end", field_type=datetime)
        if dr_end < datetime.now():
            cacheResponse(resp, cache_guid, content_type=content_type)
        return resp
    return Model.on_rest_list(request)


def getCachedResponse(cache_guid, content_type="application/json"):
    item = getCachedItem(cache_guid, content_type)
    if not item.exists:
        return None
    return _createResponse(item, cache_guid)


def getCachedItem(cache_guid, content_type="application/json"):
    url = f"s3://{settings.AWS_S3_BUCKET}/caching/{cache_guid}"
    return s3.S3Item(url, content_type)


def cacheResponse(response, cache_guid, content_type="application/json"):
    url = f"s3://{settings.AWS_S3_BUCKET}/caching/{cache_guid}"
    item = s3.S3Item(url, content_type)
    if isinstance(response, StreamingHttpResponse):
        # Access the streaming content and convert it to a file-like object
        content = io.BytesIO()
        for chunk in response.streaming_content:
            content.write(chunk)
        # Reset the file pointer to the beginning before uploading
        content.seek(0)
        item.upload(content)
        content.seek(0)
        response.streaming_content = iter(lambda: content.read(8192), b'')
    else:
        item.upload(response.content)
    return item


def _createResponse(item, cache_guid):
    # Create a StreamingHttpResponse object with the file iterator
    fp = item.download()
    fp.seek(0)
    fp.name = cache_guid
    response = StreamingHttpResponse(fp)
    # Add the appropriate content type header to the response
    # You might want to dynamically determine the content type based on the file
    response['Content-Type'] = item.content_type
    # Set the Content-Disposition header to prompt download with a filename
    response['Content-Disposition'] = f'attachment; filename="{cache_guid}"'
    return response



from django.db import models

from rest import settings
from rest import models as rm
from rest import helpers as rest_helpers

from account.models import Member
from datetime import datetime, timedelta


class Complaint(models.Model, rm.RestModel):
    class RestMeta:
        CAN_SAVE = CAN_CREATE = False
        DEFAULT_SORT = "-id"
        SEARCH_FIELDS = ["address"]
        VIEW_PERMS = ["view_logs", "view_email"]
        SEARCH_TERMS = [
            ("email", "address"),
            ("to", "address"), "source", "reason", "state",
            ("user", "user__username")]
        
        GRAPHS = {
            "default": {
                "graphs": {
                    "user": "basic"
                }
            },
            "list": {
                "graphs": {
                    "user": "basic"
                }
            }
        }
    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    user = models.ForeignKey("account.Member", related_name="emails_complaints", null=True, blank=True, default=None, on_delete=models.CASCADE)
    address = models.CharField(max_length=255, db_index=True)
    kind = models.CharField(max_length=32, db_index=True)
    reason = models.TextField(null=True, blank=True, default=None)
    user_agent = models.CharField(max_length=255, null=True, blank=True, default=None)
    source = models.CharField(max_length=255, null=True, blank=True, default=None)
    source_ip = models.CharField(max_length=64, null=True, blank=True, default=None)

    def __str__(self):
        return f"complaint: address:{self.address} reason:{self.reason}"

    @staticmethod
    def log(kind, address, reason, reporter=None, code=None, source=None, source_ip=None, user=None):
        obj = Complaint(kind=kind, address=address)
        obj.reason = reason
        obj.reporter = reporter
        obj.code = code
        obj.source = source
        obj.source_ip = source_ip
        if user is None:
            user = Member.objects.filter(email=address).last()
            # now if user has complaint we shut down right away
            if user:
                user.log("complaint", "{} complaint to {} from {}".format(kind, address, source_ip), method=kind)
                user.setProperty("notify_via", "off")
                user.log("disabled", "notifications disabled because email complaint", method="notify")
        else:
            # TODO notify support of unknown bounce
            pass
        obj.user = user
        obj.save()



from .providers import iplookup
from .providers import timezones
from .providers import location

STATES = {
    'Mississippi': 'MS', 'Oklahoma': 'OK', 'Delaware': 'DE', 'Minnesota': 'MN', 'Illinois': 'IL', 'Arkansas': 'AR',
    'New Mexico': 'NM', 'Indiana': 'IN', 'Maryland': 'MD', 'Louisiana': 'LA', 'Idaho': 'ID', 'Wyoming': 'WY',
    'Tennessee': 'TN', 'Arizona': 'AZ', 'Iowa': 'IA', 'Michigan': 'MI', 'Kansas': 'KS', 'Utah': 'UT',
    'Virginia': 'VA', 'Oregon': 'OR', 'Connecticut': 'CT', 'Montana': 'MT', 'California': 'CA',
    'Massachusetts': 'MA', 'West Virginia': 'WV', 'South Carolina': 'SC', 'New Hampshire': 'NH',
    'Wisconsin': 'WI', 'Vermont': 'VT', 'Georgia': 'GA', 'North Dakota': 'ND', 'Pennsylvania': 'PA',
    'Florida': 'FL', 'Alaska': 'AK', 'Kentucky': 'KY', 'Hawaii': 'HI', 'Nebraska': 'NE', 'Missouri': 'MO',
    'Ohio': 'OH', 'Alabama': 'AL', 'New York': 'NY', 'South Dakota': 'SD', 'Colorado': 'CO', 'New Jersey': 'NJ',
    'Washington': 'WA', 'North Carolina': 'NC', 'District of Columbia': 'DC', 'Texas': 'TX', 'Nevada': 'NV',
    'Maine': 'ME', 'Rhode Island': 'RI'}


def locateByIP(ip):
    return iplookup.getLocation(ip)


def getTimeZone(lat, lng):
    """
    returns the timezone for a lat/lng
    """
    return timezones.getTimeZone(lat, lng)


def getAddress(lat, lng):
    """
    returns the address for a lat/lng
    """
    return location.locationAt(lat, lng)


def findLocationNear(lat, lng, query=None, radius_meters=100):
    return location.findPlaces(lat, lng, query, radius_meters)


def search(address):
    return location.search(address)


def isIP(ip):
    return iplookup.isIP(ip)


def dnsToIP(dns):
    return iplookup.dnsToIP(dns)



# Generated by Django 4.1.4 on 2024-02-23 19:15

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('account', '0018_userpasskey'),
        ('location', '0003_address_postalcode_suffix'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='address',
            name='modified_by',
        ),
        migrations.AddField(
            model_name='address',
            name='group',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='addresses', to='account.group'),
        ),
        migrations.AddField(
            model_name='address',
            name='label',
            field=models.CharField(blank=True, db_index=True, default=None, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='address',
            name='member',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='addresses', to=settings.AUTH_USER_MODEL),
        ),
    ]



from rest import decorators as rd
from rest import views as rv
from .. import models as location
from datetime import datetime


@rd.urlPOST(r'^track$')
@rd.login_required
def on_new_track(request):
    """
    Creates a new track
    requires a user
    """
    kind = request.POST.get("kind", None)
    if kind is None:
        return rv.restStatus(request, False, error="requires a kind of track")

    ref_id = request.POST.get("ref_id", None)
    if ref_id is None:
        return rv.restStatus(request, False, error="requires a valid reference id")

    name = request.POST.get("name", "Generic Track")

    track = location.GeoTrack(owner=request.user, kind=kind, ref_id=int(ref_id), name=name)
    track.save()

    if "pos_lat" in request.POST:
        lat = request.POST.get("pos_lat")
        lng = request.POST.get("pos_lng")
        if lat is not None:
            elevation = request.POST.get("pos_elevation", 0.0)
            accuracy = request.POST.get("pos_accuracy", 0.0)
            altitude = request.POST.get("pos_altitude", 0.0)
            altitudeAccuracy = request.POST.get("pos_altitudeAccuracy", 0.0)
            heading = request.POST.get("pos_heading", 0.0)
            speed = request.POST.get("pos_speed", 0.0)

            pos = location.GeoPosition(
                track=track, lat=lat, lng=lng,
                elevation=elevation,
                accuracy=accuracy,
                altitude=altitude,
                altitudeAccuracy=altitudeAccuracy,
                heading=heading,
                speed=speed)
            pos.save()
    return rv.restGet(request, track)


@rd.urlPOST(r'^track/(?P<track_id>\d+)$')
@rd.login_required
def addTrackPos(request, track_id):
    try:
        track = location.GeoTrack.objects.get(pk=track_id)
    except Exception:
        return rv.restStatus(request, False, error="invalid track")

    if request.user.id != track.owner.id:
        return rv.restStatus(request, False, error="permission denied")

    if "lat" not in request.DATA or "lng" not in request.DATA:
        return rv.restStatus(request, False, error="requires valid position")

    lat = request.POST.get("lat")
    lng = request.POST.get("lng")

    elevation = request.POST.get("elevation", 0.0)
    accuracy = request.POST.get("accuracy", 0.0)
    altitude = request.POST.get("altitude", 0.0)
    altitudeAccuracy = request.POST.get("altitudeAccuracy", 0.0)
    heading = request.POST.get("heading", 0.0)
    speed = request.POST.get("speed", 0.0)

    pos = location.GeoPosition(
        track=track, lat=lat, lng=lng,
        elevation=elevation,
        accuracy=accuracy,
        altitude=altitude,
        altitudeAccuracy=altitudeAccuracy,
        heading=heading,
        speed=speed)
    pos.save()
    return rv.restStatus(request, True)


@rd.urlGET(r'^track/(?P<track_id>\d+)$')
@rd.login_required
def getTrackPositions(request, track_id):
    try:
        track = location.GeoTrack.objects.get(pk=track_id)
    except Exception:
        return rv.restStatus(request, False, error="invalid track")

    ret = location.GeoPosition.objects.filter(track=track).order_by("created")
    if "last_pos_time" in request.DATA:
        last_pos_time = datetime.fromtimestamp(float(request.DATA.get("last_pos_time", 0)))
        ret = ret.filter(created__gte=last_pos_time)
    return rv.restList(request, ret)


@rd.url(r'^geo/track$')
@rd.url(r'^geo/track/(?P<pk>\d+)$')
@rd.login_required
def rest_on_geolocation(request, pk=None):
    return location.GeoTrack.on_rest_request(request, pk)


@rd.url(r'^geo/position$')
@rd.url(r'^geo/position/(?P<pk>\d+)$')
@rd.login_required
def rest_on_geoposition(request, pk=None):
    return location.GeoPosition.on_rest_request(request, pk)




def upload(url, fp, background=False):
    raise Exception("Cannot upload to null store")

def exists(url):
    return None

def view_url(url, expires=600, is_secure=False):
    return url

def get_file(url, fp):
    raise Exception("Cannot download from null store")

def delete(url):
    pass # no-op



"""
This is where you can put handlers for running async background tasks

Task.Publish("app name", "function name")

Task.Publish("taskqueue", "on_tq_test")
"""
import time
from rest.log import getLogger
from datetime import datetime, timedelta
import random
from rest import helpers as rest_helpers
from taskqueue.models import Task


def on_tq_test(task):
    # a background task has been pushed to this call
    logger = getLogger("debug", "debug.log")
    logger.info("on_tq_test: task {}".format(task.id))
    logger.info("published at: {}".format(task.data.published_at))
    sleep_time = 20.0
    if task.data.sleep_time:
        sleep_time = task.data.sleep_time
    end_at = time.time() + sleep_time
    x = 1
    while end_at > time.time():
        if x > (9*1000000):
            x = 1
        else:
            x += x*2
            if random.randint(1,100) > 80:
                time.sleep(0.2)
    logger.info("task completed")



import boto3
from datetime import datetime, timedelta
from rest import settings
from objict import objict
from rest.extra import hostinfo


METRIC_MAP = {
    "ec2": "AWS/EC2",
    "rds": "AWS/RDS",
    "redis": "AWS/ElastiCache",
    "cpu": "CPUUtilization",
    "memory": "FreeableMemory",
    'cache_memory': "BytesUsedForCache",
    "cache_usage": "DatabaseMemoryUsagePercentage",
    "connections": "DatabaseConnections",
    "swap": "SwapUsage",
    "cons": "DatabaseConnections",
    "cache_cons": "CurrConnections",
    "max": "Maximum",
    "min": "Minimum",
    "avg": "Average",
    "ec2_cons": "EstablishedConnections"
}


def publishLocalMetrics():
    # this will publish any local metrics to cloudwatch
    instance = getLocalEC2()
    pushMetric(
        METRIC_MAP["ec2_cons"],
        hostinfo.getTcpEstablishedCount(),
        "Count", instance)


def getLocalEC2():
    """
    {
        "id": "i-0c50eec6fdf91dcaf",
        "image_id": "ami-035e0c65b8a27be8d",
        "name": "epoc.pauth.io",
        "private_ip": "172.31.49.79",
        "public_ip": "35.88.50.208",
        "security_groups": [ {
                "id": "sg-0e15d7c997b004458",
                "name": "epoc_server"
            }],
        "service": "unknown",
        "state": "running",
        "subnet_id": "subnet-e282b1ca",
        "type": "t3.medium",
        "vpc_id": "vpc-000e9565",
        "zone": "us-west-2d"
    }
    """
    return getEC2ByIP(hostinfo.getHostIP())


def getClient(region=settings.AWS_REGION, service="cloudwatch"):
    key = settings.AWS_KEY
    secret = settings.AWS_SECRET
    return boto3.client(service, aws_access_key_id=key, aws_secret_access_key=secret, region_name=region)


def pushMetric(name, value, unit="Count", instance=None):
    """Push the custom metric to CloudWatch."""
    cloudwatch = getClient(service="cloudwatch")
    if instance is None:
        instance = getLocalEC2()
    return cloudwatch.put_metric_data(
        Namespace = 'CustomMetrics',
        MetricData = [
            {
                'MetricName': name,
                'Dimensions': [
                    {
                        'Name': 'InstanceId',
                        'Value': instance.id
                    },
                ],
                'Value': value,
                'Unit': unit
            },
        ]
    )


def buildQuery(id, instance, period=300, metric="cpu", namespace="rds", stat="max"):
    dname = "DBInstanceIdentifier"
    if namespace in ["ec2", "CustomMetrics"]:
        dname = "InstanceId"
    elif namespace == "redis":
        dname = "CacheClusterId"
    mstat = dict(Namespace=METRIC_MAP.get(namespace, namespace), MetricName=metric)
    mstat["Dimensions"] = [dict(Name=dname, Value=instance)]
    return dict(Id=id, MetricStat=dict(Metric=mstat, Period=period, Stat=METRIC_MAP.get(stat, stat)))
    

def getMetrics(instances, period=300, duration_seconds=900, metric="cpu", namespace="rds", stat="max"):
    cloudwatch = getClient(service="cloudwatch")
    query = []
    id_to_instance = {}
    metric_query = METRIC_MAP.get(metric, metric)
    for i in instances:
        name = i.replace("-", "_")
        key = F"{name}_{metric}"
        id_to_instance[key] = i
        query.append(buildQuery(key, i, period, metric=metric_query, namespace=namespace, stat=stat))
    response = cloudwatch.get_metric_data(
        MetricDataQueries=query,
        StartTime=(datetime.now() - timedelta(seconds=duration_seconds)).timestamp(),
        EndTime=datetime.now().timestamp())
    if "MetricDataResults" in response:
        output = objict()
        for resp in response.get("MetricDataResults"):
            output[id_to_instance.get(resp.get("Id"), "unknown")] = resp.get("Values")
        return output
    return response


def getMetricsList(instance, namespace="rds"):
    dname = "DBInstanceIdentifier"
    if namespace == "ec2" or namespace == "CustomMetrics":
        dname = "InstanceId"
    cloudwatch = getClient(service="cloudwatch")
    response = cloudwatch.list_metrics(
        Namespace=METRIC_MAP.get(namespace, namespace),
        Dimensions=[dict(Name=dname, Value=instance)])
    metrics_list = response['Metrics']
    metrics_names = [metric['MetricName'] for metric in metrics_list]
    return metrics_names


def getAllRDS(just_ids=False, region=settings.AWS_REGION):
    # Assume a boto3 session has already been started.
    ec2_client = getClient(region=region, service="rds")
    # Use the describe_db_instances method to list all RDS instances.
    response = ec2_client.describe_db_instances()
    instances = []
    for instance in response['DBInstances']:
        instances.append(_normalizeRDS(instance))

    if not just_ids:
        return instances
    return [inst.id for inst in instances]

# Loop through the list of instances and print the instance ID.


def getAllEC2(just_ids=False, region=settings.AWS_REGION):
    """
    This function returns a list with all EC2 instances in a region.
    """
    ec2_client = getClient(region=region, service="ec2")
    response = ec2_client.describe_instances()
    # log.prettyWrite(response)
    instances_reservation = response['Reservations']
    instances_description = []
    for res in instances_reservation:
        for instance in res['Instances']:
            instances_description.append(_normalizeEC2(instance))
    if not just_ids:
        return instances_description
    return [inst.id for inst in instances_description]


def getEC2ByIP(ip):
    # retreives the local ec2 instance details by private ip
    client = getClient(service="ec2")
    filters = [dict(Name="private-ip-address", Values=[ip])]
    resp = client.describe_instances(Filters=filters)["Reservations"]
    try:
        details = resp[0]['Instances'][0]
    except Exception:
        return None
    return _normalizeEC2(details)


def _normalizeEC2(details):
    name = "unknown"
    service = "unknown"
    for tag in details["Tags"]:
        if tag["Key"] == "Name":
            name = tag["Value"]
        elif tag["Key"] == "Service":
            service = tag["Value"]
    info = objict(
        id=details["InstanceId"],
        image_id=details["ImageId"],
        type=details["InstanceType"],
        name=name,
        service=service,
        security_groups=[])

    try:
        info.zone = details["Placement"]["AvailabilityZone"]
        info.public_ip = details.get("PublicIpAddress", "")
        info.private_ip = details.get("PrivateIpAddress", "")
        info.state = details["State"]["Name"]
        info.subnet_id = details["SubnetId"]
        info.vpc_id = details["VpcId"]
    except Exception:
        pass
    
    for group in details["SecurityGroups"]:
        info.security_groups.append(dict(name=group["GroupName"], id=group["GroupId"]))
    return info


def _normalizeRDS(details):
    info = objict(
        id=details["DBInstanceIdentifier"],
        instance_class=details["DBInstanceClass"],
        engine=details["Engine"],
        engine_version=details["EngineVersion"],
        name=details.get("DBName", ""),
        endpoint=objict(),
        zone=details["AvailabilityZone"],
        multi_az=details["MultiAZ"],
        status=details["DBInstanceStatus"],
        username=details["MasterUsername"]
    )

    for key, value in details["Endpoint"].items():
        info.endpoint[key.lower()] = value
    return info


from rest import log
def getAllRedis(just_ids=False, region=settings.AWS_REGION):
    # Assume a boto3 session has already been started.
    ec2_client = getClient(region=region, service="elasticache")
    # Use the describe_db_instances method to list all RDS instances.
    response = ec2_client.describe_cache_clusters()
    instances = []
    for instance in response['CacheClusters']:
        log.pp(instance)
        instances.append(objict(
            id=instance['CacheClusterId'],
            group_id=instance['ReplicationGroupId'],
            arn=instance["ARN"],
            type=instance['CacheNodeType'],
            engine=instance['Engine'],
            engine_version=instance['EngineVersion'],
            nodes=instance['NumCacheNodes'],
            zone=instance['PreferredAvailabilityZone'],
            status=instance['CacheClusterStatus']
        ))

    if not just_ids:
        return instances
    return [inst.id for inst in instances]





ZWSID = "X1-ZWz194856f3evf_22g4o"

import requests
try:
    import xmltodict
except:
    pass

def buildPost(**kwargs):
    data = {}
    for key, value in list(kwargs.items()):
        if value:
            data[key] = value
    return data

def getPropertyValue(address=None, street1=None, street2=None, city=None, state=None, zipcode=None,**kwargs):
    url = "https://www.zillow.com/webservice/GetSearchResults.htm"
    params = {"zws-id":ZWSID}
    if address:
        if hasattr(address, "street1"):
            params["address"] = address.street1
            params["citystatezip"] = "{} {} {}".format(address.city, address.state, address.zipcode)
        elif hasattr(address, "line1"):
            params["address"] = address.line1
            params["citystatezip"] = "{} {} {}".format(address.city, address.state, address.postalcode)
        else:
            return None
    elif street1:
        params["address"] = street1
        zipcode = str(zipcode)
        if len(zipcode) > 5:
            zipcode = zipcode[:5]
        params["citystatezip"] = "{} {} {}".format(city, state, zipcode)
    else:
        return None

    print(params)
    resp = requests.get(url, params=params, timeout=15.0)
    print((resp.text))
    data = xmltodict.parse(resp.text)
    result = None
    results = data["SearchResults:searchresults"]["response"]["results"]
    if isinstance(results, list) and len(results):
        result = results[0]
    elif isinstance(results, dict):
        result = results["result"]
    if result:
        out = {}
        if "links" in result:
            out["map_url"] = result["links"].get("mapthishome", None)
        if "zestimate" in result and "amount" in result["zestimate"]:
            out["value"] = float(result["zestimate"]["amount"].get("#text", 0))
        out["zpid"] = result.get("zpid", None)
        out["address"] = result.get("address", None)
        result = out
    return result




# Generated by Django 4.0.2 on 2022-02-22 01:23

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import medialib.models
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('account', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='AccountConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Name config group', max_length=127)),
            ],
        ),
        migrations.CreateModel(
            name='CuePoint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('start', models.FloatField()),
                ('end', models.FloatField(blank=True, null=True)),
            ],
        ),
        migrations.CreateModel(
            name='MediaItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Name of item', max_length=127)),
                ('description', models.TextField(blank=True, help_text='Description of item')),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When item was created')),
                ('kind', models.CharField(choices=[('I', 'Image'), ('V', 'Video'), ('D', 'Audio'), ('L', 'Live Video'), ('A', 'Live Archive'), ('S', 'Video Segment'), ('E', 'External Link'), ('T', 'Text'), ('*', 'Unknown')], help_text='Kind of media', max_length=1)),
                ('state', models.IntegerField(choices=[(0, 'Inactive'), (1, 'Deleted'), (50, 'Archived'), (90, 'Hold'), (100, 'Initial render pending'), (101, 'Render pending'), (102, 'Rendering'), (110, 'Render failed'), (111, 'Validation failed'), (120, 'Render ending'), (200, 'Active')], help_text='Rendering state of item')),
                ('render_error', models.TextField(blank=True, help_text='Any render error messages', null=True)),
                ('group', models.ForeignKey(blank=True, default=None, help_text='Owner of item', null=True, on_delete=django.db.models.deletion.CASCADE, to='account.group')),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='MediaItemRendition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Name of rendition (copied from definition)', max_length=127)),
                ('use', models.CharField(blank=True, help_text='The intended use for this media item (copied from definition)', max_length=32, null=True)),
                ('url', models.CharField(help_text='internal url for rendition', max_length=255)),
                ('created', models.DateTimeField(auto_now_add=True, help_text='When item was created')),
                ('width', models.IntegerField(help_text='Width (if available)', null=True)),
                ('height', models.IntegerField(help_text='Height (if available)', null=True)),
                ('bytes', models.IntegerField(help_text='Size of rendition file')),
                ('kind', models.CharField(choices=[('I', 'Image'), ('V', 'Video'), ('D', 'Audio'), ('L', 'Live Stream'), ('N', 'Animated Image'), ('E', 'External Link'), ('T', 'Text'), ('*', 'Unknown')], help_text='Kind of rendition', max_length=1)),
                ('is_original', models.BooleanField(default=False, help_text='Whether is original of kind')),
                ('mediaitem', models.ForeignKey(help_text='Item this rendition is for', on_delete=django.db.models.deletion.CASCADE, related_name='renditions', to='medialib.mediaitem')),
            ],
        ),
        migrations.CreateModel(
            name='RenditionDefinition',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, help_text='Definition name', max_length=127)),
                ('description', models.TextField(blank=True, help_text='Definition description', null=True)),
                ('active', models.BooleanField(default=True, help_text='Should use for new media')),
                ('use', models.CharField(blank=True, help_text='The intended use for this rendition type', max_length=32, null=True)),
                ('depend', models.CharField(blank=True, db_index=True, help_text='Depends on (use name)', max_length=127)),
                ('configSet', models.ManyToManyField(blank=True, help_text='Additional account config set defaults', to='medialib.AccountConfig')),
            ],
            options={
                'ordering': ['preset__stage', 'id'],
            },
        ),
        migrations.CreateModel(
            name='RenditionParameter',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Parameter name', max_length=127)),
                ('description', models.TextField(blank=True, help_text='Parameter description', null=True)),
                ('required', models.BooleanField(default=False, help_text='Whether is required')),
                ('kind', models.CharField(choices=[('S', 'String'), ('I', 'Integer'), ('F', 'Float'), ('B', 'Boolean'), ('C', 'Choices')], help_text='Type of value', max_length=1)),
                ('choices', models.TextField(blank=True, help_text='Comma separated list of choices', null=True)),
            ],
        ),
        migrations.CreateModel(
            name='RenditionPreset',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Name of preset', max_length=127)),
                ('short_name', models.SlugField(help_text='short name', max_length=32, unique=True)),
                ('description', models.TextField(blank=True, help_text='Preset description', null=True)),
                ('stage', models.IntegerField(help_text='0-100: run when uploaded, >100: run later')),
                ('default_use', models.CharField(blank=True, help_text='The default intended use for this rendition types based on this preset', max_length=32, null=True)),
                ('applicable_kinds', models.CharField(help_text='Applicable rendition kinds', max_length=16)),
                ('module_name', models.SlugField(help_text='module name', max_length=32)),
                ('configurable_parameters', models.ManyToManyField(blank=True, help_text='All parameters that can be set by user', related_name='applicable_renditions', to='medialib.RenditionParameter')),
            ],
        ),
        migrations.CreateModel(
            name='RenditionSet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(db_index=True, help_text='Set name', max_length=127)),
                ('kind', models.CharField(choices=[('I', 'Image'), ('V', 'Video'), ('D', 'Audio'), ('L', 'Live Video'), ('A', 'Live Archive'), ('S', 'Video Segment'), ('E', 'External Link'), ('T', 'Text'), ('*', 'Unknown')], help_text='Kind of media this rendition set is for', max_length=1)),
                ('default_set', models.BooleanField(default=False, help_text='Whether is global default')),
                ('renditions', models.ManyToManyField(blank=True, help_text='Renditions used', to='medialib.RenditionDefinition')),
            ],
        ),
        migrations.CreateModel(
            name='RenditionSegment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('segment', models.IntegerField(help_text='Segment number')),
                ('start', models.FloatField(db_index=True, help_text='Start time offset')),
                ('end', models.FloatField(db_index=True, help_text='End time offset')),
                ('duration', models.FloatField(help_text='Segment duration')),
                ('bytes', models.IntegerField(help_text='Segment size')),
                ('url', models.CharField(help_text='internal url for rendition', max_length=255)),
                ('rendition', models.ForeignKey(help_text='Rendition this metadata belongs to', on_delete=django.db.models.deletion.CASCADE, related_name='segments', to='medialib.mediaitemrendition')),
            ],
        ),
        migrations.CreateModel(
            name='RenditionPresetParameterSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('setting', models.CharField(blank=True, help_text='Value', max_length=127, null=True)),
                ('parameter', models.ForeignKey(help_text='Parameter', on_delete=django.db.models.deletion.CASCADE, to='medialib.renditionparameter')),
                ('renditionPreset', models.ForeignKey(help_text='Preset this pair belongs to', on_delete=django.db.models.deletion.CASCADE, to='medialib.renditionpreset')),
            ],
            options={
                'unique_together': {('renditionPreset', 'parameter')},
            },
        ),
        migrations.AddField(
            model_name='renditionpreset',
            name='parameters',
            field=models.ManyToManyField(blank=True, help_text='Default values for parameters', through='medialib.RenditionPresetParameterSetting', to='medialib.RenditionParameter'),
        ),
        migrations.CreateModel(
            name='RenditionDefinitionParameterSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('setting', models.CharField(blank=True, help_text='Value', max_length=127, null=True)),
                ('parameter', models.ForeignKey(help_text='Parameter', on_delete=django.db.models.deletion.CASCADE, to='medialib.renditionparameter')),
                ('renditionDefinition', models.ForeignKey(help_text='Definition this pair belongs to', on_delete=django.db.models.deletion.CASCADE, to='medialib.renditiondefinition')),
            ],
            options={
                'unique_together': {('renditionDefinition', 'parameter')},
            },
        ),
        migrations.AddField(
            model_name='renditiondefinition',
            name='parameters',
            field=models.ManyToManyField(help_text='Configured parameters (override values only)', through='medialib.RenditionDefinitionParameterSetting', to='medialib.RenditionParameter'),
        ),
        migrations.AddField(
            model_name='renditiondefinition',
            name='preset',
            field=models.ForeignKey(help_text='Base preset', on_delete=django.db.models.deletion.CASCADE, related_name='+', to='medialib.renditionpreset'),
        ),
        migrations.CreateModel(
            name='RenderInstance',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('instance_id', models.CharField(db_index=True, help_text='Rendering engine ID', max_length=16)),
                ('state', models.CharField(choices=[('S', 'Starting'), ('I', 'Idle'), ('R', 'Rendering'), ('0', 'Stopped')], db_index=True, help_text='Current state of engine', max_length=1)),
                ('started', models.DateTimeField(auto_now_add=True, help_text='When rendering was started')),
                ('last_checkin', models.DateTimeField(default=datetime.datetime.now, help_text='time instance last checked in')),
                ('shutdown', models.DateTimeField(blank=True, help_text='Time of instance shutdown', null=True)),
                ('priority', models.IntegerField(blank=True, default=medialib.models._next_priority, help_text='Render instance priority', null=True)),
                ('message', models.TextField(blank=True, help_text='Render message', null=True)),
                ('rendering', models.ForeignKey(blank=True, help_text='Media item currently being rendered', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='render_instance', to='medialib.mediaitem')),
            ],
        ),
        migrations.CreateModel(
            name='MediaMeta',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(db_index=True, help_text='Key', max_length=127)),
                ('value', models.TextField(blank=True, help_text='Value', null=True)),
                ('rendition', models.ForeignKey(help_text='Rendition this metadata belongs to', on_delete=django.db.models.deletion.CASCADE, to='medialib.mediaitemrendition')),
            ],
        ),
        migrations.CreateModel(
            name='MediaLibrary',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='name of media library', max_length=127)),
                ('description', models.TextField(blank=True, help_text='description of media library', null=True)),
                ('allowed_kinds', models.CharField(blank=True, help_text='allowed object kinds (null=ALL)', max_length=127, null=True)),
                ('group', models.ForeignKey(blank=True, default=None, help_text='group owner of library', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='libraries', to='account.group')),
                ('owner', models.ForeignKey(blank=True, default=None, help_text='owner of library', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='libraries', to=settings.AUTH_USER_MODEL)),
                ('rendition_sets', models.ManyToManyField(blank=True, help_text='list of RenditionSets used in this library', to='medialib.RenditionSet')),
            ],
            options={
                'permissions': (('can_manage', 'Can manage medialib directly'),),
            },
        ),
        migrations.AddField(
            model_name='mediaitemrendition',
            name='rendition_definition',
            field=models.ForeignKey(help_text='The rendition type that generated this rendition', null=True, on_delete=django.db.models.deletion.CASCADE, to='medialib.renditiondefinition'),
        ),
        migrations.CreateModel(
            name='MediaItemRef',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('component', models.CharField(blank=True, db_index=True, default=None, max_length=200, null=True)),
                ('component_id', models.IntegerField(db_index=True)),
                ('media', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='references', to='medialib.mediaitem')),
            ],
            bases=(models.Model, rest.models.RestModel),
        ),
        migrations.CreateModel(
            name='MediaItemMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='medialib.mediaitem')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='mediaitem',
            name='library',
            field=models.ForeignKey(blank=True, default=None, help_text='Library this item belongs to', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='items', to='medialib.medialibrary'),
        ),
        migrations.AddField(
            model_name='mediaitem',
            name='owner',
            field=models.ForeignKey(blank=True, default=None, help_text='Owner of item', null=True, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.CreateModel(
            name='CuePointMeta',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(db_index=True, help_text='Key', max_length=127)),
                ('value', models.TextField(blank=True, help_text='Value', null=True)),
                ('cue', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='medialib.cuepoint')),
            ],
        ),
        migrations.AddField(
            model_name='cuepoint',
            name='item',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='medialib.mediaitem'),
        ),
        migrations.AddField(
            model_name='cuepoint',
            name='subitem',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='cueparent', to='medialib.mediaitem'),
        ),
        migrations.CreateModel(
            name='AccountConfigParameterSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('setting', models.CharField(blank=True, help_text='Value', max_length=127, null=True)),
                ('accountConfig', models.ForeignKey(help_text='Config this pair belongs to', on_delete=django.db.models.deletion.CASCADE, to='medialib.accountconfig')),
                ('parameter', models.ForeignKey(help_text='Parameter', on_delete=django.db.models.deletion.CASCADE, to='medialib.renditionparameter')),
            ],
            options={
                'unique_together': {('accountConfig', 'parameter')},
            },
        ),
        migrations.AddField(
            model_name='accountconfig',
            name='applicable_preset',
            field=models.ForeignKey(blank=True, help_text='Applicable preset (info only)', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='medialib.renditionpreset'),
        ),
        migrations.AddField(
            model_name='accountconfig',
            name='owner',
            field=models.ForeignKey(help_text='owner of config', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL),
        ),
        migrations.AddField(
            model_name='accountconfig',
            name='parameters',
            field=models.ManyToManyField(blank=True, help_text='Values for parameters', through='medialib.AccountConfigParameterSetting', to='medialib.RenditionParameter'),
        ),
        migrations.CreateModel(
            name='MediaItemParameterSetting',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('setting', models.CharField(blank=True, help_text='Value', max_length=127, null=True)),
                ('item', models.ForeignKey(help_text='Media Item this pair belongs to', on_delete=django.db.models.deletion.CASCADE, to='medialib.mediaitem')),
                ('parameter', models.ForeignKey(help_text='Parameter', on_delete=django.db.models.deletion.CASCADE, to='medialib.renditionparameter')),
            ],
            options={
                'unique_together': {('item', 'parameter')},
            },
        ),
    ]



from django.conf import settings as django_settings
import importlib

UNKNOWN = object()


class SettingsHelper():
    def __init__(self, root_settings, defaults=None):
        self.root = root_settings
        self.defaults = defaults
        self._app_cache = dict()

    def getAppSettings(self, app_name):
        # this will wrap a settings helper around app settings and defaults
        key = "{}_SETTINGS".format(app_name.upper())
        if key in self._app_cache:
            return self._app_cache[key]
        try:
            defaults = importlib.import_module(f"{app_name}.settings")
        except Exception:
            defaults = dict()
        self._app_cache[key] = SettingsHelper(self.get(key, dict()), defaults)
        return self._app_cache[key]

    def get(self, name, default=UNKNOWN):
        if isinstance(self.root, dict):
            res = self.root.get(name, UNKNOWN)
        else:
            res = getattr(self.root, name, UNKNOWN)
        if res == UNKNOWN:
            return self.getDefault(name, default)
        return res

    def getDefault(self, name, default=None):
        if isinstance(self.defaults, dict):
            return self.defaults.get(name, default)
        return getattr(self.defaults, name, default)

    def __getattr__(self, name):
        """Access settings as an attribute."""
        return self.get(name, None)

    def __getitem__(self, key):
        """Access settings as a dictionary key."""
        return self.get(key)


settings = SettingsHelper(django_settings)



from rest import net
from rest import helpers as rh
import requests
REQ_TIMEOUT = 15.0


def REQUEST(task):
    method = task.fname.upper()
    url = task.data.url
    params = None
    if "?" in url:
        url, params = net.parse_params(url)
    try:
        resp = net.REQUEST(method, url, params=params,
                           data=task.data.data, raw_response=True,
                           verify=False, timeout=REQ_TIMEOUT)
        task.log(resp.text, kind="response")
        if resp.status_code == 200:
            return True
        task.reason = f"invalid response of {resp.status_code}"
    except requests.Timeout:
        task.reason = "request timed out"
        task.log("request timed out", kind="error")
    except Exception as err:
        task.reason = str(err)
        task.log_exception(err)
    return False



from django.db import models
from django.db.models import Sum, Max, Count, Q
from django.db.models.functions import Trunc
from rest import helpers as rh
from rest import models as rm
from rest import settings
from datetime import datetime, timedelta
from rest import datem as date_util
from objict import objict
from . import utils
from .eod import EODMetrics

# THIS IS IN DAYS
METRICS_EXPIRE_HOURLY = settings.get("METRICS_EXPIRE_HOURLY", 90)
METRICS_EXPIRE_DAILY = settings.get("METRICS_EXPIRE_DAILY", 360)
METRICS_EXPIRE_WEEKLY = settings.get("METRICS_EXPIRE_WEEKLY", 360)

def metric(
    slug, keys, data, min_granularity="hourly",
    group=None, date=None, timezone=None, slug_append=None,
    max_granularity=None
):
    # keys is a ordered list of keys to map to k1,k2,etc
    # data is a dict of key/values
    uuid_key = generate_uuid(slug, group, slug_append)
    granularities = utils.granularities(min_granularity, max_granularity)
    date = normalize_date(date, timezone, group)

    uuids = utils.build_keys(
        uuid_key, date,
        min_granularity=min_granularity,
        max_granularity=max_granularity)

    for granularity, key in zip(granularities, uuids):
        # DO not change slug to uuid_key so we can filter by group using group field
        expires = None
        if granularity == "hourly":
            expires = datetime.now() + timedelta(days=METRICS_EXPIRE_HOURLY)
        elif granularity == "daily":
            expires = datetime.now() + timedelta(days=METRICS_EXPIRE_DAILY)
        elif granularity == "weekly":
            expires = datetime.now() + timedelta(days=METRICS_EXPIRE_WEEKLY)
        m, created = Metrics.objects.get_or_create(
            uuid=key, 
            defaults=dict(
                granularity=granularity, slug=slug, expires=expires,
                group=group, start=utils.date_for_granulatiry(date, granularity)))
        m.updateMetrics(keys, data, created)


def gauge(slug, keys, data, granularity="daily", group=None, date=None, 
          timezone=None, slug_append=None, max_granularity=None):
    # guage does not accumulate but just stores the data like a cache
    # if calledf on the same time period it will just update the current numbers
    if max_granularity is None:
        max_granularity = granularity
    uuid_key = generate_uuid(slug, group, slug_append)
    granularities = utils.granularities(granularity, max_granularity)
    date = normalize_date(date, timezone, group)
    uuids = utils.build_keys(
        uuid_key, date,
        min_granularity=granularity,
        max_granularity=max_granularity)

    for gran, key in zip(granularities, uuids):
        # DO not change slug to uuid_key so we can filter by group using group field
        expires = None
        if gran == "hourly":
            expires = datetime.now() + timedelta(days=METRICS_EXPIRE_HOURLY)
        elif gran == "daily":
            expires = datetime.now() + timedelta(days=METRICS_EXPIRE_DAILY)
        m, created = Metrics.objects.get_or_create(
            uuid=key, 
            defaults=dict(
                granularity=gran, slug=slug, expires=expires,
                group=group, start=utils.date_for_granulatiry(date, gran)))
        m.setMetrics(keys, data, created)


def normalize_date(date=None, timezone=None, group=None):
    if date is None:
        date = utils.datetime.utcnow()
    if timezone is not None:
        date = date_util.convertToLocalTime(timezone, date)
    elif group:
        date = group.getLocalTime(date)
    elif settings.METRICS_TIMEZONE:
        date = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, date)
    return date

def get_qset(slug, granularity, start=None, end=None,
             group=-1, starts_with=False, ends_with=None):
    if start is None:
        start = utils.datetime.utcnow()
    if group is not None and group != -1:
        start = group.getLocalTime(start)
    elif settings.METRICS_TIMEZONE:
        start = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, start)
  
    start = utils.date_for_granulatiry(start, granularity)
    if end is None:
        if granularity == "hourly":
            end = start + timedelta(minutes=5)
        else:
            end = start + timedelta(hours=12)
    elif group != -1 and end:
        end = group.getLocalTime(end)
    elif settings.METRICS_TIMEZONE:
        end = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, end)
    q = {
        "granularity": granularity,
        "start__gte": start,
        "start__lte": end
    }
    if group != -1:
        q["group"] = group
    if starts_with:
        q["slug__startswith"] = slug
    else:
        q["slug"] = slug
    if ends_with:
        q["slug__endswith"] = ends_with
    return Metrics.objects.filter(**q)


def get_totals(slug, keys, granularity, start=None, end=None, group=None):
    if start is None:
        start = utils.datetime.utcnow()
    if group:
        start = group.getLocalTime(start)
    elif settings.METRICS_TIMEZONE:
        start = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, start)
  
    start = utils.date_for_granulatiry(start, granularity)
    if end is None:
        end = start + timedelta(minutes=5)
    elif group and end:
        end = group.getLocalTime(end)
    elif settings.METRICS_TIMEZONE:
        end = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, end)
    
    qset = Metrics.objects.filter(
        slug=slug, granularity=granularity,
        group=group, start__gte=start, start__lte=end)
    vkeys = [f"v{i}" for i in range(1, len(keys)+1)]
    sums = rh.getSum(qset, *vkeys)
    out = objict()
    i = 1
    for k in keys:
        out[k] = sums[f"v{i}"]
        i += 1
    return objict(slug=slug, granularity=granularity, start=start, end=end, values=out)


def get_metric(slug, granularity, start, group=None):
    if start is None:
        start = utils.datetime.utcnow()
    if group:
        start = group.getLocalTime(start)
    elif settings.METRICS_TIMEZONE:
        start = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, start)
  
    start = utils.date_for_granulatiry(start, granularity)
    qset = Metrics.objects.filter(
        slug=slug, granularity=granularity, start__gte=start, start__lte=start, group=group)
    m = qset.last()
    if m is None:
        return objict()
    return m.getMetrics()


def get_metrics(slug, granularity, start, end=None, group=None, samples=None):
    """
    returns data ready for Chart.js
    'periods': ['y:2012', 'y:2013', 'y:2014']
    'data': [
      {
        'slug': 'bar',
        'values': [1, 2, 3]
      },
      {
        'slug': 'foo',
        'values': [4, 5, 6]
      },
    ]
    """
    # convert to local time of group or metrics
    has_start = start is not None
    start = utils.getLocalDate(start, group)
    end = utils.getLocalDate(end, group)
    # convert to granularit start and end
    if not has_start:
        # this will find a new start based on the granularity and sample size
        start = utils.start_by_granularity(granularity, start, samples=samples)
    # set the correct end for this granularity
    end = utils.date_for_granulatiry(end, granularity)
    qset = Metrics.objects.filter(
        slug=slug, granularity=granularity, start__gte=start, start__lte=end, group=group)
    # now we want to get the keys from the most recent object (latest keys if changed)
    keys = None
    obj = qset.last()
    if obj is None:
        return objict(periods=[], data={})
    values = obj.getMetrics()
    keys = values.keys()
    period_values = []
    for obj in qset.order_by("uuid"):
        values = obj.getMetrics()
        pvals = objict(uuid=obj.uuid, period=utils.slug_to_label(obj.uuid), values=values)
        period_values.append(pvals)
    periods = get_chart_periods(slug, granularity, start, end)
    # now we assume that our raw metrics are sorted correctly
    data = dict()
    if keys is None:
        return objict(periods=periods, data=data)
    for k in keys:
        data[k] = []
    # this logic is not perfect
    # it will find the first match first the last match
    for period in periods:
        result = next((d for d in period_values if d.get("period") == period), None)
        if result is None:
            for k in keys:
                data[k].append(0)
            continue
        for k in keys: 
            data[k].append(result["values"].get(k, 0))
        period_values.remove(result)
    return objict(periods=periods, data=data)


def get_chart_periods(slug, granularity, start, end=None, group=None):
    if end is None:
        start, end = get_adjusted_date_range(granularity, start, end, group=group)
    periods = []
    for date in utils.date_range(granularity, start, end):
        period = utils.slug_to_label(utils.build_keys(slug, date, granularity)[0])
        periods.append(period)
    if granularity != "weekly":
        periods.reverse()
    return periods   


def get_adjusted_date_range(granularity, start, end, group=None, samples=None):
    if start is None:
        if group is not None:
            start = group.getLocalTime(start)
        else:
            start = utils.getLocalDate(None)
        end = start
        start = utils.start_by_granularity(granularity, start, samples=samples)
    elif group:
        start = group.getLocalTime(start)
    elif settings.METRICS_TIMEZONE:
        start = date_util.convertToLocalTime(settings.METRICS_TIMEZONE, start)
    if end is None:
        end = utils.datetime.utcnow()
    end = utils.date_for_granulatiry(end, granularity)
    return start, end


def get_category_metrics(category, since=None, granularity="daily"):
    """Create/Increment a metric."""
    # r = get_r()
    # slugs = [utils.to_string(s) for s in list(r.category_slugs(category))]
    # print(slugs)
    # return r.get_metric_history_chart_data(slugs, since, granularity)
    return None


def generate_uuid(slug, group, slug_append=None):
    if group is not None and slug_append is not None:
        return f"{slug}.{slug_append}.{group.pk}"
    elif group is not None:
        return f"{slug}.{group.pk}"
    return slug


class Metrics(models.Model, rm.RestModel):
    class RestMeta:
        QUERY_FIELDS = ["group__kind", "all_fields"]
        GRAPHS = {
            "detailed": {
                "fields": [
                    "id", "created", "start",
                    "granularity", "uuid",
                    "slug"
                ],
                "extra": [("getMetrics", "metrics")],
                "graphs": {
                    "group": "basic"
                }
            }
        }

    created = models.DateTimeField(auto_now_add=True, editable=True)
    # timeframe of metrics
    start = models.DateTimeField(db_index=True)
    # auto delete objects older then this
    expires = models.DateTimeField(db_index=True, null=True, default=None)
    # granularity: year, month, week, day, hour
    granularity = models.CharField(max_length=64, db_index=True)
    # unique uuid of metric
    uuid = models.SlugField(max_length=124, unique=True)
    # kind/slug of metric
    slug = models.SlugField(max_length=124, db_index=True)
    # allow to group metrics by a group
    group = models.ForeignKey("account.Group", related_name="+", on_delete=models.CASCADE, null=True, default=None)

    # now we create a set of k/v 
    k1 = models.SlugField(max_length=64, null=True, default=None)
    v1 = models.BigIntegerField(default=0)

    k2 = models.SlugField(max_length=64, null=True, default=None)
    v2 = models.BigIntegerField(default=0)

    k3 = models.SlugField(max_length=64, null=True, default=None)
    v3 = models.BigIntegerField(default=0)

    k4 = models.SlugField(max_length=64, null=True, default=None)
    v4 = models.BigIntegerField(default=0)

    k5 = models.SlugField(max_length=64, null=True, default=None)
    v5 = models.BigIntegerField(default=0)

    k6 = models.SlugField(max_length=64, null=True, default=None)
    v6 = models.BigIntegerField(default=0)

    k7 = models.SlugField(max_length=64, null=True, default=None)
    v7 = models.BigIntegerField(default=0)

    k8 = models.SlugField(max_length=64, null=True, default=None)
    v8 = models.BigIntegerField(default=0)

    k9 = models.SlugField(max_length=64, null=True, default=None)
    v9 = models.BigIntegerField(default=0)

    k10 = models.SlugField(max_length=64, null=True, default=None)
    v10 = models.BigIntegerField(default=0)

    k11 = models.SlugField(max_length=64, null=True, default=None)
    v11 = models.BigIntegerField(default=0)

    k12 = models.SlugField(max_length=64, null=True, default=None)
    v12 = models.BigIntegerField(default=0)

    k13 = models.SlugField(max_length=64, null=True, default=None)
    v13 = models.BigIntegerField(default=0)

    k14 = models.SlugField(max_length=64, null=True, default=None)
    v14 = models.BigIntegerField(default=0)

    def getMetrics(self):
        metrics = objict()
        for i in range(1, 15):
            key = getattr(self, f"k{i}", None)
            if key is None:
                return metrics
            metrics[key] = getattr(self, f"v{i}", 0)
        return metrics

    def setMetrics(self, keys, data, update_keys=False):
        self.updateMetrics(keys, data, update_keys, False)

    def updateMetrics(self, keys, data, update_keys=False, add_values=True):
        params = {}
        index = 0
        if not update_keys:
            # this fixes bug when adding new keys after creation
            i = len(keys)
            update_keys = getattr(self, f"k{i}", None) != keys[-1]
        for key in keys:
            if isinstance(data, list):
                v = data[index]
            else:
                v = data[key]
            index += 1
            vkey = f"v{index}"
            if add_values:
                params[vkey] = models.F(vkey) + v
            else:
                params[vkey] = v
            if update_keys:
                params[f"k{index}"] = key
        Metrics.objects.filter(pk=self.pk).update(**params)
